#include <cassert>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <omp.h>
#include <random>
#include <stdint.h>
#include <string>

#if defined(WIN32)
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#include "stopwatch_win32.h"
#undef min
#undef max
#else
#include <unistd.h>
#include "stopwatch_std.h"
#endif

typedef char Level[16][16];

struct Task
{
	Level level;
	int correctSolution;
};

#include "solution_reference.h"

#include "solution_david.h"
#include "solution_helmut.h"
#include "solution_schrompf.h"
#include "solution_stef.h"

// Erzeugt zufällige Levels basierend auf Level-Templates.
void generateLevels()
{
	std::ifstream levelTemplatesFile("level_templates.txt");
	std::ofstream levelsFile("levels.txt");
	std::mt19937 rng(42 * 1337);
	
	int levelTemplateNo = 1;
	while (!levelTemplatesFile.eof())
	{
		// Level-Vorlage zeilenweise lesen.
		Level levelTemplate;
		for (int y = 0; y < 16; ++y)
		{
			std::string line;
			levelTemplatesFile >> line;
			for (int x = 0; x < 16; ++x) levelTemplate[x][y] = line[x];
		}
		
		// Die Null einlesen.
		int unused = 0;
		levelTemplatesFile >> unused;
		
		int numSolvableLevelsGenerated = 0;
		
		while (numSolvableLevelsGenerated < 1000)
		{
			//#pragma omp parallel
			{
				Level level;
				for (int x = 0; x < 16; ++x)
					for (int y = 0; y < 16; ++y)
						level[x][y] = levelTemplate[x][y];
				
				for (int i = 0; i < 6; ++i)
				{
					const char fieldToPlace = "rgbRGB"[i];
					while (true)
					{
						const int x = 1 + rng() % 15;
						const int y = 1 + rng() % 15;
						if (level[x][y] == '.')
						{
							level[x][y] = fieldToPlace;
							break;
						}
					}
				}
				
				int correctSolution = Reference::numMovesToSolve(level);
				if (correctSolution != -1 && correctSolution <= 255)
				{
					//#pragma omp critical
					{
						++numSolvableLevelsGenerated;
						
						for (int y = 0; y < 16; ++y)
						{
							std::string line;
							for (int x = 0; x < 16; ++x) line.append(1, level[x][y]);
							levelsFile << line << std::endl;
						}
						
						levelsFile << correctSolution << std::endl;
						
						std::cout << "Level " << levelTemplateNo << " / " << numSolvableLevelsGenerated << " erzeugt!" << std::endl;
					}
				}
			}
		}
		
		++levelTemplateNo;
	}
}

void evaluateSolution(const std::string& name, int (* p_solution)(const Level&), const std::vector<Task>& tasks)
{
	std::cout << "Teste Funktion von " << name << " ..." << std::endl;
	
	// Mehrere Durchläufe durch alle Levels. Der beste zählt!
	double minTimeNeeded = 0;
	Stopwatch stopwatch;
	for (int i = 0; i < 10; ++i)
	{
		std::cout << "Durchlauf " << (i + 1) << " ... " << std::flush;
		
		stopwatch.reset();
		int levelNo = 1;
		
		for (std::vector<Task>::const_iterator it = tasks.begin(); it != tasks.end(); ++it, ++levelNo)
		{
			const int answer = (*p_solution)(it->level);
			if (answer != it->correctSolution)
			{
				std::cout << std::endl << "Fehler bei Level " << levelNo << ": Antwort war " << answer << " statt " << it->correctSolution << "!" << std::endl << std::endl;
				return;
			}
		}
		
		const double timeNeeded = stopwatch();
		std::cout << "Zeit: " << timeNeeded << " Sekunden" << std::endl;
		
		if (timeNeeded < minTimeNeeded || minTimeNeeded == 0) minTimeNeeded = timeNeeded;
	}
	
	std::cout << "Bestzeit: " << minTimeNeeded << " Sekunden" << std::endl << std::endl;
}

void evaluateSolutions()
{
	std::ifstream levelsFile("levels.txt");
	std::vector<Task> tasks;
	
	std::cout << "Lade Levels ..." << std::endl;
	
	while (!levelsFile.eof())
	{
		Task task;
		
		// Level zeilenweise lesen.
		for (int y = 0; y < 16; ++y)
		{
			std::string line;
			levelsFile >> line;
			for (int x = 0; x < 16; ++x) task.level[x][y] = line[x];
		}
		
		// Die korrekte Lösung einlesen.
		levelsFile >> task.correctSolution;
		
		tasks.push_back(task);
	}
	
	std::cout << tasks.size() << " Levels geladen!" << std::endl << std::endl;
	std::cout << "Starten mit Taste!" << std::endl;
	std::cin.get();
	
	evaluateSolution("David", &David::numMovesToSolve, tasks);
	evaluateSolution("Helmut", &Helmut::numMovesToSolve, tasks);
	evaluateSolution("Schrompf", &Schrompf::numMovesToSolve, tasks);
	evaluateSolution("Stef", &Stef::numMovesToSolve, tasks);
}

int main()
{
#if defined(WIN32)
	{
		// Priorität des Prozesses und des Threads hochsetzen.
		HANDLE process = GetCurrentProcess();
		HANDLE thread = GetCurrentThread();
		SetPriorityClass(process, REALTIME_PRIORITY_CLASS);
		SetThreadPriority(thread, THREAD_PRIORITY_TIME_CRITICAL);
		
		// Der Thread soll nur auf einem Prozessor laufen.
		// Ansonsten kann es Probleme bei der Zeitmessung mit QueryPerformanceCounter geben.
		// Wir benutzen den "letzten" Prozessor des Systems.
		DWORD threadMask = SetThreadAffinityMask(thread, 1);
		int highestBit = 0;
		for (int i = 0; i < 32; ++i) if (threadMask & (1 << i)) highestBit = i;
		SetThreadAffinityMask(thread, 1 << highestBit);
	}
#endif

	// generateLevels();
	
	evaluateSolutions();
	
	return 0;
}