namespace David
{
int numMovesToSolve(const Level& level)
{
	/* 
	 * Lösungsverfahren: Bidirektionale Suche
	 * --------------------------------------
	 * 
	 * Es wird ein Weg vom Start- zum Zielzustand gesucht und "parallel" dazu
	 * ein Weg vom Ziel- zum Startzustand. Beide Suchen sind einfache Breitensuchen.
	 * Wenn bei der Vorwärtssuche ein Zustand erreicht wird, der bereits bei der
	 * Rückwärtssuche erreicht wurde (oder umgekehrt), ist ein kürzester Weg gefunden.
	 * Nach jeder Suche werden die Anzahlen der noch zu untersuchenden Knoten beider Suchen
	 * verglichen. Es wird dann mit der Suchrichtung fortgefahren, die weniger offene Knoten hat,
	 * um beide Suchrichtungen auszubalancieren.
	 * Vorteil der bidirektionalen Suche: die Suchtiefe halbiert sich!
	 * Die Laufzeit einer Breitensuche wächst exponentiell mit der Suchtiefe.
	 * Zwei Suchen mit halber Tiefe sind also im Allgemeinen schneller als eine mit voller Tiefe.
	 * 
	 * 
	 * Rückwärtssuche
	 * --------------
	 * 
	 * Bei der Rückwärtssuche müssen zu einem Zustand alle möglichen Vorgängerzustände
	 * aufgelistet werden. Dazu muss man für jede Spielfigur alle Felder finden, von denen sie
	 * gekommen sein kann.
	 * Wenn das Feld rechts von der Figur blockiert ist (durch eine Mauer oder eine andere Figur),
	 * dann kann sie von links gekommen sein. Analoges gilt für die anderen Richtungen.
	 * Beispiel:
	 * 
	 * ###########                                     ###########   ###########   ###########
	 * #.........#                                     #.........#   #.........#   #.........#
	 * #...r######                                     #..r.######   #.r..######   #r...######
	 * #....######   hat folgende Vorgängerzustände:   #....######   #....######   #....######
	 * #.........#                                     #.........#   #.........#   #.........#
	 * #.........#                                     #.........#   #.........#   #.........#
	 * ###########                                     ###########   ###########   ###########
	 * 
	 * 
	 * Kodierung der Zustände
	 * ----------------------
	 * 
	 * Für die Vor- und Rückwärtssuche müssen die noch zu untersuchenden Zustände in einer
	 * Warteschlange gespeichert werden. Dabei wird ein Zustand in einen uint32_t-Wert kodiert.
	 * Die x- und y-Koordinaten jeder Spielfigur nehmen jeweils 4 Bits ein, das macht
	 * insgesamt 24 Bits für die Positionen der Spielfiguren.
	 * In den verbleibenden 8 Bits wird noch die Bewegung gespeichert (Spielfigur und Achse),
	 * über die der Zustand gefunden wurde. Bei der Untersuchung des Zustands weiß man nun,
	 * dass eine erneute Bewegung dieser Spielfigur entlang derselben Achse unnötig ist,
	 * weil alle daraus folgenden Zustände bereits entdeckt wurden.
	 * Im Beispiel von oben müsste also bei der Untersuchung des letzten Zustands nicht wieder
	 * die Bewegung der roten Spielfigur nach rechts in Betracht gezogen werden.
	 * 
	 */
	
	// "Statisches struct" für das Bit-Array der Vorwärtssuche.
	struct BitArrayFwd
	{
		// Testet, ob ein Bit in einem Bit-Array gesetzt ist.
		static inline bool test(const uint32_t* p_data, uint32_t bit)
		{
			return p_data[(bit >> 4) & 0xFFFFFFFE] & (1 << (bit & 31));
		}
		
		// Setzt ein Bit in einem Bit-Array.
		static inline void set(uint32_t* p_data, uint32_t bit)
		{
			p_data[(bit >> 4) & 0xFFFFFFFE] |= 1 << (bit & 31);
		}
		
		// Testet, ob ein Bit in einem Bit-Array gesetzt ist.
		// Wenn nicht, dann wird es gesetzt.
		// Der alte Wert des Bits wird zurückgeliefert.
		static inline bool testAndSet(uint32_t* p_data, uint32_t bit)
		{
			uint32_t& n = p_data[(bit >> 4) & 0xFFFFFFFE];
			const uint32_t mask = 1 << (bit & 31);
			if (n & mask) return true;
			n += mask;
			return false;
		}
	};
	
	// Fast dasselbe nochmal für die Rückwärtssuche.
	struct BitArrayBwd
	{
		// Testet, ob ein Bit in einem Bit-Array gesetzt ist.
		static inline bool test(const uint32_t* p_data, uint32_t bit)
		{
			return p_data[(bit >> 4) | 1] & (1 << (bit & 31));
		}
		
		// Setzt ein Bit in einem Bit-Array.
		static inline void set(uint32_t* p_data, uint32_t bit)
		{
			p_data[(bit >> 4) | 1] |= 1 << (bit & 31);
		}
		
		// Testet, ob ein Bit in einem Bit-Array gesetzt ist.
		// Wenn nicht, dann wird es gesetzt.
		// Der alte Wert des Bits wird zurückgeliefert.
		static inline bool testAndSet(uint32_t* p_data, uint32_t bit)
		{
			uint32_t& n = p_data[(bit >> 4) | 1];
			const uint32_t mask = 1 << (bit & 31);
			if (n & mask) return true;
			n += mask;
			return false;
		}
	};
	
	// Makros zum Setzen der x-/y-Position einer Spielfigur in einem Zustand.
	// Sie setzen voraus, dass die entsprechenden Bits bereits genullt wurden.
#define SET_X_FAST(STATE, PLAYER, X) { (STATE) |= (X) << (8 * (PLAYER)); }
#define SET_Y_FAST(STATE, PLAYER, Y) { (STATE) |= (Y) << (4 + 8 * (PLAYER)); }
	
	// Spielsteine und Zielfelder finden.
	// Dabei auch die Begehbarkeitskarte initialisieren.
	// Das Array ist mit Absicht 16x16 statt 14x14 (einfachere Berechnung der Speicheradressen).
	uint32_t startState = 0;
	uint32_t goalState = 0;
	bool isFree[16][16];
	int numFreeFields = 0;
	for (int x = 0; x < 14; ++x)
	{
		for (int y = 0; y < 14; ++y)
		{
			const char field = level[x + 1][y + 1];
			if (field == '#') isFree[x][y] = false;
			else
			{
				isFree[x][y] = true;
				++numFreeFields;
				switch (field)
				{
				case 'r':
					SET_X_FAST(startState, 0, x)
					SET_Y_FAST(startState, 0, y)
					break;
				case 'g':
					SET_X_FAST(startState, 1, x)
					SET_Y_FAST(startState, 1, y)
					break;
				case 'b':
					SET_X_FAST(startState, 2, x)
					SET_Y_FAST(startState, 2, y)
					break;
				case 'R':
					SET_X_FAST(goalState, 0, x)
					SET_Y_FAST(goalState, 0, y)
					break;
				case 'G':
					SET_X_FAST(goalState, 1, x)
					SET_Y_FAST(goalState, 1, y)
					break;
				case 'B':
					SET_X_FAST(goalState, 2, x)
					SET_Y_FAST(goalState, 2, y)
					break;
				}
			}
		}
	}
	
	// Wir legen ein 2D-Array an, mit dem man schnell herausfinden kann,
	// bis zu welcher Stelle man rutscht, wenn man sich an einer bestimmten Stelle
	// in eine bestimmte Richtung bewegt. Dies berücksichtigt aber nur die Mauern,
	// nicht die anderen Spielfiguren.
	// Ein Eintrag enthält alle 4 Werte für links, rechts, oben und unten (jeweils 4 Bits).
	uint32_t stopPosLookup[16][16];
	for (int x = 0; x < 14; ++x)
	{
		for (int y = 0; y < 14; ++y)
		{
			if (!isFree[x][y]) continue;
			
			// Bewegung nach links bis zur nächsten Wand.
			int pos = x;
			while (pos > 0 && isFree[pos - 1][y]) --pos;
			stopPosLookup[x][y] = pos;
			
			// Bewegung nach rechts bis zur nächsten Wand.
			pos = x;
			while (pos < 13 && isFree[pos + 1][y]) ++pos;
			stopPosLookup[x][y] |= (pos << 4);
			
			// Bewegung nach oben bis zur nächsten Wand.
			pos = y;
			while (pos > 0 && isFree[x][pos - 1]) --pos;
			stopPosLookup[x][y] |= (pos << 8);
			
			// Bewegung nach unten bis zur nächsten Wand.
			pos = y;
			while (pos < 13 && isFree[x][pos + 1]) ++pos;
			stopPosLookup[x][y] |= (pos << 12);
		}
	}
	
	// Warteschlangen für die noch zu untersuchenden Zustände anlegen.
	// Die Warteschlangen werden mittels "zirkulärem Array" implementiert,
	// dessen Größe zur Laufzeit konstant bleibt.
	// Interessanterweise ist die Variante mit einer Zweierpotenzgröße
	// und Indizes langsamer als die mit beliebiger Größe und Zeigern.
 	// Ich mache die Warteschlangen groß genug, dass theoretisch alle Zustände hineinpassen.
	// Die Anzahl theoretisch möglicher Zustände berechnet sich aus der Anzahl
	// der freien Felder im Level.
	const int queueSize = numFreeFields * (numFreeFields - 1) * (numFreeFields - 2);
	uint32_t* const p_queues = new uint32_t[2 * queueSize];
	uint32_t* const p_fwdQueue = p_queues;
	uint32_t* const p_bwdQueue = p_queues + queueSize;
	const uint32_t* const p_fwdQueueEnd = p_fwdQueue + queueSize;
	const uint32_t* const p_bwdQueueEnd = p_bwdQueue + queueSize;
	const uint32_t* p_fwdQueueFront = p_fwdQueue;
	const uint32_t* p_bwdQueueFront = p_bwdQueue;
	uint32_t* p_fwdQueueBack = p_fwdQueue;
	uint32_t* p_bwdQueueBack = p_bwdQueue;
	
	// In den "gesehen"-Bit-Arrays merken wir uns, welche Zustände wir schon bei der
	// Vorwärts- und Rückwärtssuche gesehen haben, damit sie nicht doppelt
	// untersucht werden und wir wissen, wann wir einen kürzesten Weg gefunden haben.
	// Die Bit-Arrays müssen am Anfang mit Nullen initialisiert werden.
	// Hierbei hat sich "calloc" als schneller herausgestellt als "new []" und "memset".
	// Außerdem werden die Bit-Arrays "verzahnt" angelegt, d.h. der erste Wert dient
	// der Vorwärtssuche, der zweite der Rückwärtssuche, und so weiter.
	// Da immer kurz hintereinander derselbe Bit-Index im Vorwärts- und Rückwärts-Array
	// angesprochen wird, verbessert dies (zumindest theoretisch) die Cache-Nutzung.
	const int seenSize = 0xDDDCDB / 32 + 1;
	uint32_t* const p_seen = static_cast<uint32_t*>(calloc(2 * seenSize, 4));
	
	// Start- und Endzustand in die jeweilge Warteschlange schreiben
	// und als "gesehen" markieren.
	*(p_fwdQueueBack++) = startState;
	BitArrayFwd::set(p_seen, startState);
	(*p_bwdQueueBack++) = goalState;
	BitArrayBwd::set(p_seen, goalState);
	
	// In diesen Variablen merken wir uns, wie teuer die Zustände sind, die wir gerade betrachten.
	// Es ist also nicht nötig, die Kosten in jedem Zustand zu speichern.
	int fwdCurrentCosts = 0;
	int bwdCurrentCosts = 0;
	
	// Hier merken wir uns, ab welcher Stelle der Warteschlange die Zustände um eins teurer werden.
	const uint32_t* p_fwdIncCosts = p_fwdQueueBack;
	const uint32_t* p_bwdIncCosts = p_bwdQueueBack;
	
	while (true)
	{
		// Wir schauen uns nun die beiden Warteschlangen an und entscheiden
		// durch den Vergleich ihrer Größe, ob wir als Nächstes vorwärts oder rückwärts suchen.
		// Es wird immer dort weiter gesucht, wo die wenigsten Knoten offen sind.
		// Die Rückwärtssuche kann schnell "explodieren"!
		if (   (p_fwdQueueBack + (p_fwdQueueBack < p_fwdQueueFront ? queueSize : 0) - p_fwdQueueFront)
			<= (p_bwdQueueBack + (p_bwdQueueBack < p_bwdQueueFront ? queueSize : 0) - p_bwdQueueFront))
		{
			// Vorwärtssuche!
			while (p_fwdQueueFront != p_fwdIncCosts)
			{
				// Zustand aus der Warteschlange holen und den Hint extrahieren.
				const uint32_t state = *p_fwdQueueFront & 0x00FFFFFF;
				const uint32_t stateHint = *p_fwdQueueFront & 0xFF000000;
				if (++p_fwdQueueFront == p_fwdQueueEnd) p_fwdQueueFront = p_fwdQueue;
				
#define INIT_VARS \
	/* Hier werden die Positionen der Spielfiguren aus dem Zustand geladen. */ \
	/* Außerdem ermitteln wir die Stopppositionen aller Figuren und Richtungen. */ \
	/* Die Stopppositionen berücksichtigen aber noch nicht die anderen Figuren. */ \
	const int posX0 = state & 15; \
	const int posY0 = (state >> 4) & 15; \
	const int posX1 = (state >> 8) & 15; \
	const int posY1 = (state >> 12) & 15; \
	const int posX2 = (state >> 16) & 15; \
	const int posY2 = state >> 20; \
	const int stopPos0 = stopPosLookup[posX0][posY0]; \
	const int stopPos1 = stopPosLookup[posX1][posY1]; \
	const int stopPos2 = stopPosLookup[posX2][posY2]; \
	int stopLeft0 = stopPos0 & 15; \
	int stopRight0 = (stopPos0 >> 4) & 15; \
	int stopUp0 = (stopPos0 >> 8) & 15; \
	int stopDown0 = stopPos0 >> 12; \
	int stopLeft1 = stopPos1 & 15; \
	int stopRight1 = (stopPos1 >> 4) & 15; \
	int stopUp1 = (stopPos1 >> 8) & 15; \
	int stopDown1 = stopPos1 >> 12; \
	int stopLeft2 = stopPos2 & 15; \
	int stopRight2 = (stopPos2 >> 4) & 15; \
	int stopUp2 = (stopPos2 >> 8) & 15; \
	int stopDown2 = stopPos2 >> 12;
				
				INIT_VARS
				
#define PROCESS_COLLISIONS(A, B) \
	/* Hier wird geprüft, ob die Figuren A und B kollidieren können. */ \
	/* Ihre Stopppositionen werden entsprechend angepasst. */ \
	{ \
		if (posX##A == posX##B) \
		{ \
			if (posY##B < posY##A) { if (posY##B >= stopUp##A) stopUp##A = posY##B + 1, stopDown##B = posY##A - 1; } \
			else if (posY##B <= stopDown##A) stopDown##A = posY##B - 1, stopUp##B = posY##A + 1; \
		} \
		else if (posY##A == posY##B) \
		{ \
			if (posX##B < posX##A) { if (posX##B >= stopLeft##A) stopLeft##A = posX##B + 1, stopRight##B = posX##A - 1; } \
			else if (posX##B <= stopRight##A) stopRight##A = posX##B - 1, stopLeft##B = posX##A + 1; \
		} \
	}
				
#define PROCESS_PLAYER_FWD(N, O1, O2, MASK_X, MASK_Y, SHIFT_X, SHIFT_Y, HINT_X, HINT_Y) \
	{ \
		if (stateHint != HINT_X) \
		{ \
			/* Bewegung nach links bis zur Stoppposition. */ \
			/* Die Bitmaske MASK_X setzt die Bits der x-Position der Spielfigur auf 0. */ \
			uint32_t newState = state & MASK_X; \
			newState |= stopLeft##N << SHIFT_X; \
			if (!BitArrayFwd::testAndSet(p_seen, newState)) \
			{ \
				/* Wenn dieser Zustand für die Vorwärtssuche neu ist, */ \
				/* aber bei der Rückwärtssuche schon gefunden wurde, dann sind wir fertig. */ \
				/* Achtung: böses "goto"! */ \
				if (BitArrayBwd::test(p_seen, newState)) goto done; \
				/* Der neue Zustand wird im nächsten Schritt weiter verfolgt. */ \
				/* Mit dem Hint merken wir uns, dass die letzte Bewegung mit dieser Figur */ \
				/* auf der x-Achse stattfand. */ \
				*p_fwdQueueBack = newState + HINT_X; \
				if (++p_fwdQueueBack == p_fwdQueueEnd) p_fwdQueueBack = p_fwdQueue; \
			} \
			\
			/* Bewegung nach rechts. */ \
			newState &= MASK_X; \
			newState |= stopRight##N << SHIFT_X; \
			if (!BitArrayFwd::testAndSet(p_seen, newState)) \
			{ \
				if (BitArrayBwd::test(p_seen, newState)) goto done; \
				*p_fwdQueueBack = newState + HINT_X; \
				if (++p_fwdQueueBack == p_fwdQueueEnd) p_fwdQueueBack = p_fwdQueue; \
			} \
		} \
		\
		if (stateHint != HINT_Y) \
		{ \
			/* Bewegung nach oben. */ \
			uint32_t newState = state & MASK_Y; \
			newState |= stopUp##N << SHIFT_Y; \
			if (!BitArrayFwd::testAndSet(p_seen, newState)) \
			{ \
				if (BitArrayBwd::test(p_seen, newState)) goto done; \
				*p_fwdQueueBack = newState + HINT_Y; \
				if (++p_fwdQueueBack == p_fwdQueueEnd) p_fwdQueueBack = p_fwdQueue; \
			} \
			\
			/* Bewegung nach unten. */ \
			newState &= MASK_Y; \
			newState |= stopDown##N << SHIFT_Y; \
			if (!BitArrayFwd::testAndSet(p_seen, newState)) \
			{ \
				if (BitArrayBwd::test(p_seen, newState)) goto done; \
				*p_fwdQueueBack = newState + HINT_Y; \
				if (++p_fwdQueueBack == p_fwdQueueEnd) p_fwdQueueBack = p_fwdQueue; \
			} \
		} \
	}
				
				PROCESS_COLLISIONS(0, 1)
				PROCESS_COLLISIONS(0, 2)
				PROCESS_COLLISIONS(1, 2)
				
				PROCESS_PLAYER_FWD(0, 1, 2, 0x00FFFFF0, 0x00FFFF0F, 0, 4, 0x01000000, 0x02000000)
				PROCESS_PLAYER_FWD(1, 2, 0, 0x00FFF0FF, 0x00FF0FFF, 8, 12, 0x03000000, 0x04000000)
				PROCESS_PLAYER_FWD(2, 0, 1, 0x00F0FFFF, 0x000FFFFF, 16, 20, 0x05000000, 0x06000000)
			}
			
			++fwdCurrentCosts;
			p_fwdIncCosts = p_fwdQueueBack;
		}
		else
		{
			// Rückwärtssuche!
			while (p_bwdQueueFront != p_bwdIncCosts)
			{
				// Zustand aus der Warteschlange holen und den Hint extrahieren.
				const uint32_t state = *p_bwdQueueFront & 0x00FFFFFF;
				const uint32_t stateHint = *p_bwdQueueFront & 0xFF000000;
				if (++p_bwdQueueFront == p_bwdQueueEnd) p_bwdQueueFront = p_bwdQueue;
				
				INIT_VARS
				
				PROCESS_COLLISIONS(0, 1)
				PROCESS_COLLISIONS(0, 2)
				PROCESS_COLLISIONS(1, 2)
				
#define PROCESS_PLAYER_BWD(N, MASK_X, MASK_Y, SHIFT_X, SHIFT_Y, HINT_X, HINT_Y) \
	{ \
		if (stateHint != HINT_X) \
		{ \
			/* Kann die Spielfigur im vorherigen Zug von links gekommen sein? */ \
			/* Das geht nur, wenn rechts von ihr eine Mauer oder eine andere Figur ist. */ \
			if (stopRight##N == posX##N) \
			{ \
				uint32_t newState = state & MASK_X; \
				uint32_t stopState = newState; \
				newState |= stopLeft##N << SHIFT_X; \
				stopState |= posX##N << SHIFT_X; \
				for (; newState < stopState; newState += 1 << (8 * N)) \
				{ \
					if (!BitArrayBwd::testAndSet(p_seen, newState)) \
					{ \
						/* Wenn dieser Zustand bei der Vorwärtssuche schon gefunden wurde, */ \
						/* dann sind wir jetzt fertig. */ \
						if (BitArrayFwd::test(p_seen, newState)) goto done; \
						*p_bwdQueueBack = newState + HINT_X; \
						if (++p_bwdQueueBack == p_bwdQueueEnd) p_bwdQueueBack = p_bwdQueue; \
					} \
				} \
			} \
			/* Kann sie von rechts gekommen sein? */ \
			/* Prinzipiell geht das nur, wenn sie nicht von links gekommen sein kann (darum "else"). */ \
			else if (stopLeft##N == posX##N) \
			{ \
				uint32_t newState = state & MASK_X; \
				uint32_t stopState = newState; \
				newState |= stopRight##N << SHIFT_X; \
				stopState |= posX##N << SHIFT_X; \
				for (; newState > stopState; newState -= 1 << (8 * N)) \
				{ \
					if (!BitArrayBwd::testAndSet(p_seen, newState)) \
					{ \
						if (BitArrayFwd::test(p_seen, newState)) goto done; \
						*p_bwdQueueBack = newState + HINT_X; \
						if (++p_bwdQueueBack == p_bwdQueueEnd) p_bwdQueueBack = p_bwdQueue; \
					} \
				} \
			} \
		} \
		\
		if (stateHint != HINT_Y) \
		{ \
			/* Kann sie von oben gekommen sein? */ \
			if (stopDown##N == posY##N) \
			{ \
				uint32_t newState = state & MASK_Y; \
				uint32_t stopState = newState; \
				newState |= stopUp##N << SHIFT_Y; \
				stopState |= posY##N << SHIFT_Y; \
				for (; newState < stopState; newState += 1 << (4 + 8 * N)) \
				{ \
					if (!BitArrayBwd::testAndSet(p_seen, newState)) \
					{ \
						if (BitArrayFwd::test(p_seen, newState)) goto done; \
						*p_bwdQueueBack = newState + HINT_Y; \
						if (++p_bwdQueueBack == p_bwdQueueEnd) p_bwdQueueBack = p_bwdQueue; \
					} \
				} \
			} \
			/* Kann sie von unten gekommen sein? */ \
			else if (stopUp##N == posY##N) \
			{ \
				uint32_t newState = state & MASK_Y; \
				uint32_t stopState = newState; \
				newState |= stopDown##N << SHIFT_Y; \
				stopState |= posY##N << SHIFT_Y; \
				for (; newState > stopState; newState -= 1 << (4 + 8 * N)) \
				{ \
					if (!BitArrayBwd::testAndSet(p_seen, newState)) \
					{ \
						if (BitArrayFwd::test(p_seen, newState)) goto done; \
						*p_bwdQueueBack = newState + HINT_Y; \
						if (++p_bwdQueueBack == p_bwdQueueEnd) p_bwdQueueBack = p_bwdQueue; \
					} \
				} \
			} \
		} \
	}
				
				PROCESS_PLAYER_BWD(0, 0x00FFFFF0, 0x00FFFF0F, 0, 4, 0x01000000, 0x02000000)
				PROCESS_PLAYER_BWD(1, 0x00FFF0FF, 0x00FF0FFF, 8, 12, 0x03000000, 0x04000000)
				PROCESS_PLAYER_BWD(2, 0x00F0FFFF, 0x000FFFFF, 16, 20, 0x05000000, 0x06000000)
			}
			
			++bwdCurrentCosts;
			p_bwdIncCosts = p_bwdQueueBack;
		}
	}
	
done:
	delete[] p_queues;
	free(p_seen);
	return fwdCurrentCosts + bwdCurrentCosts + 1;
	
#undef SET_X_FAST
#undef SET_Y_FAST
#undef INIT_VARS
#undef PROCESS_COLLISIONS
#undef PROCESS_PLAYER_FWD
#undef PROCESS_PLAYER_BWD
}
}