namespace Schrompf
{
/// Gibt eine Zahl zurück, bei der die untersten x Bits gesetzt sind und alle anderen gelöscht
inline uint32_t MachUntenEins( size_t pAnzahl) { return (1u << pAnzahl) - 1u; }
/// Gibt eine Zahl zurück, bei der alle bis auf die untersten x Bits gesetzt sind
inline uint32_t MachObenEins( size_t pAnzahl) { return 0xfffffffeu << pAnzahl; }
/// Ermittelt den Index des höchsten gesetzten Bit (von 0 an zählend) aus der gegebenen Zahl. Hardcore-fixiert auf 16Bits
inline size_t FindeHoechstesBit( uint32_t pZahl) 
{ 
  unsigned int v = pZahl;  // 32-bit value to find the log2 of 
  static const unsigned char LogTable256[256] = 
  {
#define LT(n) n, n, n, n, n, n, n, n, n, n, n, n, n, n, n, n
    -1, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3,
    LT(4), LT(5), LT(5), LT(6), LT(6), LT(6), LT(6),
    LT(7), LT(7), LT(7), LT(7), LT(7), LT(7), LT(7), LT(7)
  };
#undef LT

  register unsigned int t; // temporaries
  return (t = v >> 8) ? 8 + LogTable256[t] : LogTable256[v];
}
/// Ermittelt den Index des niedrigsten gesetzten Bit (von 0 an zählend) aus der gegebenen Zahl
inline size_t FindeNiedrigstesBit( uint32_t pZahl) 
{ 
  unsigned int v = pZahl;     // 32-bit word input to count zero bits on right
  unsigned int c = 1;
  if ((v & 0xff) == 0) 
  {  
    v >>= 8;  
    c += 8;
  }
  if ((v & 0xf) == 0) 
  {  
    v >>= 4;
    c += 4;
  }
  if ((v & 0x3) == 0) 
  {  
    v >>= 2;
    c += 2;
  }
  c -= v & 0x1;

  return c;
}

// kleine eigene Array-Klasse, die nicht bei jedem Einfügen auf Vergrößerung prüft
class Feld
{
  uint32_t* feld;
  uint32_t* ende;
  size_t kapazitaet;

public:
  Feld() { feld = ende = nullptr; kapazitaet = 0; }
  ~Feld() { delete [] feld; }

  void reserve( size_t groesse) { feld = new uint32_t[groesse]; ende = feld; kapazitaet = groesse; }
  void push_back( uint32_t wert) { *ende++ = wert; }
  void erase( uint32_t* start, uint32_t* schluss) { ende = feld; }
  const uint32_t* cbegin() const { return feld; }
  uint32_t* begin() { return feld; }
  const uint32_t* cend() const { return ende; }
  uint32_t* end() { return ende; }
  bool empty() const { return feld == ende; }
};

// kleine Helferklasse zum Speichern der schon erreichten Positionen
class BesetztFeld
{
  uint32_t* feld;

public:
  BesetztFeld( size_t groesse) { feld = new uint32_t[(groesse+31)/32]; memset( feld, 0, (groesse+7)/8); }
  ~BesetztFeld() { delete [] feld; }

  bool Test( size_t index) const { return (feld[index >> 5] & (1 << (index & 31))) != 0; }
  void Set( size_t index) { feld[index >> 5] |= (1 << (index & 31)); }
};

class KollKarte
{
  uint16_t koll[16];

public:
  KollKarte() { koll[0] = koll[15] = 0xffff; std::fill( koll+1, koll+15, 0x8001); }
  KollKarte( const KollKarte& k) { memcpy( koll, k.koll, 32); }

  void Set( size_t px, size_t py) { koll[py] |= (1 << px); }
  void Entfernen( size_t px, size_t py) { koll[py] ^= (1 << px); }
  bool Test( size_t px, size_t py) { return (koll[py] & (1 << px)) != 0; }
  uint16_t GetVorZeile( size_t px, size_t py) const { return koll[py] & ((1 << px)-1); }
  uint16_t GetNachZeile( size_t px, size_t py) const { return koll[py] >> (px+1); }
  const uint16_t* GetZeiger() const { return koll; }
};
/// Fügt die gegebene Positionskombination zur Folgepos-Sammlung hinzu, falls sie noch nicht besucht wurde
inline void AddFolgePos( Feld& pFolgeSammlung, uint32_t pFolgePos, BesetztFeld& pBesuchtePositionen)
{
  if( pBesuchtePositionen.Test( pFolgePos & 0xffffff) )
    return;
  pBesuchtePositionen.Set( pFolgePos & 0xffffff);
  pFolgeSammlung.push_back( pFolgePos);
}

// Die zu programmierende Funktion:
// Gegeben ist ein garantiert in höchstens 255 Zügen lösbares Level.
// Gesucht, als Rückgabewert, ist die Anzahl der Züge der optimalen Lösung (Mindestanzahl).
int numMovesToSolve(const Level& level)
{
  KollKarte karteQuer, karteHoch;
  // Startpositionen und Zielpositionen als Bitkombination zu je 4 Bit der Reihenfolge rx, ry, gx, gy, bx, by
  uint32_t startpos, zielpos;

  // Sammlung möglicher Folgezüge mit Zielposition: 24Bit Position, oberste 8 Bit Zug dahin
  // und da bedeutet: 0 links oder rechts, 1 hoch oder runter, |0 rot, |2 grün, |4 blau
  Feld folgezuege1, folgezuege2;
  folgezuege1.reserve( 3000000); folgezuege2.reserve( 3000000);

  // true, falls wir die Spielstein-Positionskombination schon erprobt haben
  BesetztFeld besuchtePositionen( 256*256*256);

  {
    size_t rx, ry, gx, gy, bx, by;
    size_t zrx, zry, zgx, zgy, zbx, zby;
    for( size_t y = 1; y < 15; ++y )
    {
      for( size_t x = 1; x < 15; ++x )
      {
        switch( level[x][y] )
        { 
        case 'R': zrx = x; zry = y; break;
        case 'G': zgx = x; zgy = y; break;
        case 'B': zbx = x; zby = y; break;
        case 'r': rx = x; ry = y; break;
        case 'g': gx = x; gy = y; break;
        case 'b': bx = x; by = y; break;
        case '#': karteQuer.Set( x, y); karteHoch.Set( y, x); break;
        }
      }
    }

    startpos = rx | (ry << 4) | (gx << 8) | (gy << 12) | (bx << 16) | (by << 20);
    zielpos = zrx | (zry << 4) | (zgx << 8) | (zgy << 12) | (zbx << 16) | (zby << 20);
  }


  // Startposition mit ungültigem Vorher-Zug ablegen
  folgezuege1.push_back( startpos | (6 << 24));

  // jetzt jeweils alle möglichen Positionen prüfen und deren gültige und sinnvolle Folgezüge sammeln
  int anzZuege = 0;
  while( true ) 
  {    
    // alle Positionen aus diesem Stapel
    auto& positionen = ((anzZuege & 1) == 1) ? folgezuege2 : folgezuege1;
    // ergeben Folgepositionen in diesen Stapel
    auto& folgepos = ((anzZuege & 1) == 1) ? folgezuege1 : folgezuege2;
    folgepos.erase( folgepos.begin(), folgepos.end());

    for( auto it = positionen.cbegin(); it != positionen.cend(); ++it )
    {
      uint32_t ap = *it & 0xffffffu;

      if( ap == zielpos )
        return anzZuege;

      size_t rx = ap & 15, ry = ap >> 4 & 15;
      size_t gx = ap >> 8 & 15, gy = ap >> 12 & 15;
      size_t bx = ap >> 16 & 15, by = ap >> 20 & 15;
      size_t vorzug = *it >> 24;

      // Kollisionsmap mit den drei Steinen besetzen 
      KollKarte lokalKarteQuer( karteQuer), lokalKarteHoch( karteHoch);
      lokalKarteQuer.Set( rx, ry); lokalKarteHoch.Set( ry, rx);
      lokalKarteQuer.Set( gx, gy); lokalKarteHoch.Set( gy, gx);
      lokalKarteQuer.Set( bx, by); lokalKarteHoch.Set( by, bx);

      // Folgezüge durchtesten
      // Rot nach links und rechts. Ergibt nur Sinn, wenn Rot nicht im letzten Zug schon horizontal gezogen wurde
      if( vorzug != 0 )
      {
        size_t nrx = FindeHoechstesBit( lokalKarteQuer.GetVorZeile( rx, ry)) + 1;
        if( nrx != rx ) AddFolgePos( folgepos, ap & 0xfffff0 | nrx | 0x0000000, besuchtePositionen);

        // Rot nach rechts.
        nrx = FindeNiedrigstesBit( lokalKarteQuer.GetNachZeile( rx, ry));
        if( nrx != 0 ) AddFolgePos( folgepos, ap & 0xfffff0 | (rx+nrx) | 0x0000000, besuchtePositionen);
      }

      // Rot hoch oder runter
      if( vorzug != 1 )
      {
        size_t nry = FindeHoechstesBit( lokalKarteHoch.GetVorZeile( ry, rx)) + 1;
        if( nry != ry ) AddFolgePos( folgepos, ap & 0xffff0f | (nry << 4) | 0x1000000, besuchtePositionen);
        nry = FindeNiedrigstesBit( lokalKarteHoch.GetNachZeile( ry, rx));
        if( nry != 0 ) AddFolgePos( folgepos, ap & 0xffff0f | ((ry+nry) << 4) | 0x1000000, besuchtePositionen);
      }

      // Grün nach links oder rechts
      if( vorzug != 2 )
      {
        size_t ngx = FindeHoechstesBit( lokalKarteQuer.GetVorZeile( gx, gy)) + 1;
        if( ngx != gx ) AddFolgePos( folgepos, ap & 0xfff0ff | (ngx << 8) | 0x2000000, besuchtePositionen);
        ngx = FindeNiedrigstesBit( lokalKarteQuer.GetNachZeile( gx, gy));
        if( ngx != 0 ) AddFolgePos( folgepos, ap & 0xfff0ff | ((gx+ngx) << 8) | 0x2000000, besuchtePositionen);
      }
      // Grün hoch oder runter
      if( vorzug != 3 )
      {
        size_t ngy = FindeHoechstesBit( lokalKarteHoch.GetVorZeile( gy, gx)) + 1;
        if( ngy != gy ) AddFolgePos( folgepos, ap & 0xff0fff | (ngy << 12) | 0x3000000, besuchtePositionen);
        ngy = FindeNiedrigstesBit( lokalKarteHoch.GetNachZeile( gy, gx));
        if( ngy != 0 ) AddFolgePos( folgepos, ap & 0xff0fff | ((gy+ngy) << 12) | 0x3000000, besuchtePositionen);
      }

      // Blau nach links oder rechts
      if( vorzug != 4 )
      {
        size_t nbx = FindeHoechstesBit( lokalKarteQuer.GetVorZeile( bx, by)) + 1;
        if( nbx != bx ) AddFolgePos( folgepos, ap & 0xf0ffff | (nbx << 16) | 0x4000000, besuchtePositionen);
        nbx = FindeNiedrigstesBit( lokalKarteQuer.GetNachZeile( bx, by));
        if( nbx != 0 ) AddFolgePos( folgepos, ap & 0xf0ffff | ((bx + nbx) << 16) | 0x4000000, besuchtePositionen);
      }
      // Blau hoch oder runter
      if( vorzug != 5 )
      {
        size_t nby = FindeHoechstesBit( lokalKarteHoch.GetVorZeile( by, bx)) + 1;
        if( nby != by ) AddFolgePos( folgepos, ap & 0x0fffff | (nby << 20) | 0x5000000, besuchtePositionen);
        nby = FindeNiedrigstesBit( lokalKarteHoch.GetNachZeile( by, bx));
        if( nby != 0 ) AddFolgePos( folgepos, ap & 0x0fffff | ((by + nby) << 20) | 0x5000000, besuchtePositionen);
      }
    }

    ++anzZuege;

    // falls das Array der Folgepositionen leer ist, ist was richtig faul
    if( folgepos.empty() )
      break;
    // bei mehr als x Schritten ebenso
    if( anzZuege > 50 )
      break;
  }

  // soweit sollte es nie kommen
  return 0;
}
}